/*******************************************************************
 *  Copyright (c) 1994-2002 Jetico, Inc., Finland
 *  All rights reserved.
 *
 *  File:          wipe.c
 *
 *  Description:   implementation of BestCrypt wipe utility
 *
 *  Author:        Vitaliy Zolotarev
 *
 *  Created:       20-Nov-1999
 *
 *  Revision:      $Id: wipe.c,v 1.8 2002/12/11 08:22:09 crypt Rel-1.2-3 $
 *
 *
 *******************************************************************/


#include <stdio.h>
#include <sys/types.h>
#include <dirent.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <string.h>
#include <utime.h>
#include <stdlib.h>
#include <sys/ioctl.h>
//#include <sys/vfs.h>
#include <time.h>
#include <ctype.h>
#include <sys/mount.h>
#include <time.h>
#include "wipe.h"
#include "sha1random.h"


char wipe_c[]="$Id: wipe.c,v 1.8 2002/12/11 08:22:09 crypt Rel-1.2-3 $";

#ifndef NAME_MAX
    #ifdef MAXNAMELEN
        #define NAME_MAX MAXNAMELEN
    #endif 
#endif 

#ifndef PATH_MAX
    #ifdef MAXPATHLEN
        #define PATH_MAX MAXPATHLEN
    #endif 
#endif 


#ifndef S_BLKSIZE
    #define S_BLKSIZE   512
#endif    

unsigned long long int total_size_of_slacks=0;



/*ask_y_n(char *string) */
/*asks for yes or no */
/*return value: */
/*TRUE if 'y' or 'a' entered */
/*FALSE in other cases */
int ask_y_n(char *string)
{
    char c;
    fprintf(stderr,string);
    c = tolower(fgetc(stdin));
    if ( '\n' == c ) return FALSE;
    while ( '\n' != fgetc(stdin) );
    if ( 'y' == c ) return TRUE;
    if ( 'a' == c )
    {
        o_interactive=FALSE;
        return TRUE;
    }

    return FALSE;
}


/*  wipe_filename(char *filename) */
/*  wipe filename by renaming */
/*  return value: */
/*  ERROR   on error */
/*  OK      on success */
int wipe_filename(char *filename)
{
    struct utimbuf utb;

    if ( o_dont_delete )return OK;

    if ( o_verbose )
    {
        fprintf (stdout, "Removing file      %s ",filename);
    }
    if ( o_dont_wipe_fn )
    {
        if ( OK != unlink(filename) )
        {
            if ( o_verbose )
                fprintf (stderr, "\t - %s\n",strerror(errno));
            else
                fprintf (stderr, "Removing %s - %s\n",filename,strerror(errno));
            return ERROR;
        }
        
        if ( o_verbose )
        {
            fprintf (stdout, "\t - Done\n");
        }
        return OK;
    }

    truncate(filename,0);         /* set filesize to 0 */
    utb.actime=utb.modtime=0;
    utime(filename,&utb);         /* set access and modification time to 0 */

    if ( OK != unlink(filename) )
    {
        if ( o_verbose )
            fprintf (stderr, "\t - %s\n",strerror(errno));
        else
            fprintf (stderr, "Removing %s - %s\n",filename,strerror(errno));
        return ERROR;
    }
    if ( o_verbose )
    {
        fprintf (stdout, "\t - Done\n");
    }
    return OK;
}


/*fill_buff(char *buff, size_t size, int n) */
/*fill buffrer buff coresponding pass number n */
int fill_buff(char *buff, size_t size, int n, off_t s)
{
    int i,j;
    time_t t;
    if ( n == o_pas_num - 1 )
    {
        if ( s && o_use_buff ) return OK;
        if ( o_use_rand )
            for ( i = 0; i < size; i++ ) buff[i]=(char)(rand() & 0xff);
        else
        {
            t=time(NULL);
            SHA1RandomGenerate((unsigned char*)buff, size, (unsigned char*)&t, sizeof(t));
        }
        return OK;
    }
    if ( o_use_DoD )
    {
        if (QPASSES != o_pas_num)
        {
            if ( s ) return OK;
            j = buff[0];
            buff[0] = (char)(rand() & 0xff);
            for ( i = 0; i < size; i++ )
                if ( 1 == (n % 2) ) buff[i] = ~j;
                else buff[i] = buff[0];
            return OK;
        }else
        {
            if ( 0 == dod_pass[n].len )
            {
                if ( s && o_use_buff ) return OK;
                if ( o_use_rand )
                    for ( i = 0; i < size; i++ ) buff[i]=(char)(rand() & 0xff);
                else
                {
                    t=time(NULL);
                    SHA1RandomGenerate((unsigned char*)buff, size, (unsigned char*)&t, sizeof(t));
                }
            }
            else if ( 0 == s )
                for ( i = 0, j = 0; i < size; i++ )
                    buff[i] = dod_pass[n].pat[j++ % dod_pass[n].len];

            return OK;
        }
    }
    if ( 0 == pg_pass[n].len )
    {
        if ( s && o_use_buff ) return OK;
        if ( o_use_rand )
            for ( i = 0; i < size; i++ ) buff[i]=(char)(rand() & 0xff);
        else
        {
            t=time(NULL);
            SHA1RandomGenerate((unsigned char*)buff, size, (unsigned char*)&t, sizeof(t));
        }
    }
    else if ( 0 == s )
        for ( i = 0, j = 0; i < size; i++ )
            buff[i] = pg_pass[n].pat[j++ % pg_pass[n].len];

    return OK;
}

/*off_t find_size_by_reading(int fd) */
/*return value: */
/*0       on filure */
/*size    on success */
long long int find_size_by_reading(int fd)
{
    off_t limit,pos,prev,b_pos=0;
    char buf[64];
    int i;

    for ( limit=0x7f,i=sizeof(limit);--i; )
        limit = (limit <<8)|0xff;

    pos=prev=limit;
    while ( pos != b_pos && pos <= limit )
    {
/*        fprintf(stdout,"pos=%20ld   b_pos=%12ld   buf=%x\n",(long int)pos,(long int)b_pos,buf[0]&0xff); */
        if ( pos != lseek(fd,pos,SEEK_SET) )
        {
            fprintf(stderr,"lseek error: %s\n",strerror(errno));
            return 0;
        }
        if ( 1 != read(fd,buf,1) )
        {
            prev=pos;
            pos=(pos+b_pos)>>1;
        }
        else
        {
            b_pos=pos;
            pos=(pos+prev)>>1;
        }
    }

    if ( pos >= limit )
        return 0;

    return lseek(fd,0,SEEK_CUR);
}

int wipe_file_slack (char *filename)
{
    char *buff;
    int fd;
    unsigned int   i;
    ssize_t size_of_slack;
    struct stat st;
    uid_t uid;
    struct utimbuf time_buf;

    if ( ERROR == lstat(filename,&st) )
    {
        fprintf(stderr,"Stating file %s: %s  \t - Skiping\n",filename,strerror(errno));
        return ERROR;
    }


    if ( !S_ISREG(st.st_mode))
        return OK;

    uid = geteuid();

    if (0 != uid && st.st_uid != uid)
    {
        fprintf(stderr,"You are not owner of %s  \t - Skiping\n",filename);
        return OK;
    }
    if ( o_verbose )
    {
        fprintf (stdout, "Wiping slack of      %s",filename);
    }

    size_of_slack = st.st_blksize - (st.st_size % st.st_blksize);
    if (st.st_blksize == size_of_slack)
    {
        if ( o_verbose )
            fprintf(stdout," - Done \n");
        return OK;
    }

    
    if ( OK != access(filename,W_OK|R_OK) )
    {
        if ( OK != chmod(filename,S_IRUSR|S_IWUSR|st.st_mode) )
        {
            fprintf(stderr,"Chmod file %s: %s  \t - Skiping\n",filename,strerror(errno));
            return ERROR;
        }
    }
    
    buff = malloc(st.st_blksize);
    if (NULL == buff)
    {
        fprintf(stderr,"Allocating buffer error: %s  \t - Skiping\n",strerror(errno));
        return OK;
    }

#ifdef O_SYNC
    fd = open(filename,O_RDWR|O_SYNC);
#else
    fd = open(filename,O_RDWR);
#endif
    if ( ERROR == fd )
    {
        fprintf(stderr,"Opening file %s: %s  \t - Skiping\n",filename,strerror(errno));
        free(buff);
        return ERROR;
    }

    
    for ( i=0; i < o_pas_num; i++ )
    {
        if (ERROR == lseek(fd,st.st_size,SEEK_SET))
        {
            fprintf(stderr,"Seek to end of file %s: %s  \t - Skiping\n",filename,strerror(errno));
            ftruncate(fd,st.st_size);
            close(fd);
            free(buff);
            return ERROR;
        }
        fill_buff(buff, sizeof(buff), i,0);
        if (ERROR == write(fd,(const void *)buff,size_of_slack))
        {
            if (o_verbose)
                fprintf(stderr,"\n");
            fprintf(stderr,"Writing to %s: %s\n",filename,strerror(errno));
            ftruncate(fd,st.st_size);
            close(fd);
            free(buff);
            return ERROR;
        }
        
#ifndef O_SYNC
        (void)fsync(fd);
#endif
    }

    if ( o_verbose )
        fprintf(stdout," - Done \n");
    
    ftruncate(fd,st.st_size);
    
    close(fd);
    
    free(buff);
    
    total_size_of_slacks +=size_of_slack;
    if ( OK != chmod(filename,st.st_mode) )
    {
        fprintf(stderr,"Restore mode of file %s: %s  \n",filename,strerror(errno));
        return ERROR;
    }
    
    time_buf.actime  = st.st_atime; 
    time_buf.modtime = st.st_mtime;
    if ( OK != utime(filename,&time_buf) )
    {
        fprintf(stderr,"Restore times of file %s: %s  \n",filename,strerror(errno));
        return ERROR;
    }
    
    return OK;
}


/* wipe_file (char *filename) */
/* return value: */
/*  ERROR on error */
/*  OK    on success */
int wipe_file (char *filename)
{
    char buff[BUFFSIZE];
    int fd;
    unsigned int   i,op,pr,opr;
    ssize_t j;
    time_t t,ot,bt;
    long long int s=0;
    long long int os=0;
    long long int size;
    struct stat st;

    if (o_wipe_slacks)
        return wipe_file_slack(filename);

    if ( ERROR == lstat(filename,&st) )
    {
        fprintf(stderr,"Stating file %s: %s  \t - Skiping\n",filename,strerror(errno));
        return ERROR;
    }
    if ( S_ISREG(st.st_mode))
    {
        size = (long long int)st.st_blocks * S_BLKSIZE;
    }else
        size = st.st_size;

    if ( o_interactive )
    {
        fprintf(stderr,"Wipe %s ",filename);
        if ( !ask_y_n("(y/[n]/a)?") ) return ERROR;
    }

    if ( o_verbose )
    {
        if ( S_ISBLK(st.st_mode) || S_ISCHR(st.st_mode) )
            fprintf (stdout, "Wiping device      %s\n",filename);
        else
            fprintf (stdout, "Wiping   file      %s\n",filename);
    }

    if ( 0 == st.st_size && !S_ISBLK(st.st_mode) && !S_ISCHR(st.st_mode))
        return OK;

    if ( OK != access(filename,W_OK|R_OK) )
    {
        if ( !o_force )
        {
            fprintf(stderr,"Wiping file %s: %s  \t - Skiping\n",filename,strerror(errno));
            return ERROR;
        }
        if ( OK != chmod(filename,S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH) )
        {
            fprintf(stderr,"Chmod file %s: %s  \t - Skiping\n",filename,strerror(errno));
            return ERROR;
        }
    }

#ifdef O_SYNC
    fd = open(filename,O_RDWR|O_SYNC);
#else
    fd = open(filename,O_RDWR);
#endif
    if ( ERROR == fd )
    {
        fprintf(stderr,"Opening file %s: %s  \t - Skiping\n",filename,strerror(errno));
        return ERROR;
    }

    if ( S_ISBLK(st.st_mode) || S_ISCHR(st.st_mode) )
    {

#ifdef BLKGETSIZE
        long l;
        if ( !ioctl(fd,BLKGETSIZE,&l) )
            size=(long long)l*S_BLKSIZE;
#endif    

        if ( 0 == size || ERROR == size )
            size=lseek(fd,0,SEEK_END);
        if ( 0 == size || ERROR == size )
        {
            size = find_size_by_reading(fd);
        }
        if ( 0 == size || ERROR == size )
            for ( size=0x7f,i=sizeof(size);--i; )
                size = (size <<8)|0xff;
    }
    
    for ( i=0, op=0 ; i < o_pas_num; i++ )
    {
        if ( s && (S_ISBLK(st.st_mode) || S_ISCHR(st.st_mode)) ) size=s;
        s=os=0;
        pr=opr=0;
        ot=bt=time(NULL);
        lseek(fd,0,SEEK_SET);
        while ( s < size )
        {
            j = size - s < BUFFSIZE ? size - s : BUFFSIZE;
            fill_buff(buff, BUFFSIZE, i,s);

            if ( ERROR == (j=write(fd,(const void *)buff,j)) )
            {
                if ( ENOSPC == errno && (S_ISBLK(st.st_mode) || S_ISCHR(st.st_mode)) ) break;
                fprintf(stderr,"Writing to %s: %s\n",filename,strerror(errno));
                close(fd);
                return ERROR;
            }

            s+=j;
#ifndef O_SYNC
            (void)fsync(fd);
#endif
            if ( o_verbose )
            {
                t=time(NULL);
                pr=(int)(s*100/size);
                if ( t != ot )
                {
                    if ( op == i ) fprintf(stdout,"\rpass %2d/%d :%10ld/%ld Kb (%3d%%)   Rate: %ld Kb/s          ",i+1,o_pas_num,(long int)(s>>10),(long int)(size>>10),pr,(long int)((s/(long long int)(t-bt))>>10));
                    else         fprintf(stdout,"\rpass %2d/%d :%10ld/%ld Kb (%3d%%)                           ",i+1,o_pas_num,(long int)(s>>10),(long int)(size>>10),pr);
                    ot=t;
                    op=i;
                    opr=pr;
                }
                else if ( pr != opr )
                {
                    if ( bt != t ) fprintf(stdout,"\rpass %2d/%d :%10ld/%ld Kb (%3d%%)   Rate: %ld Kb/s          ",i+1,o_pas_num,(long int)(s>>10),(long int)(size>>10),pr,(long int)((s/(long long int)(t-bt))>>10));
                    else         fprintf(stdout,"\rpass %2d/%d :%10ld/%ld Kb (%3d%%)  ",i+1,o_pas_num,(long int)(s>>10),(long int)(size>>10),pr);
                    opr=pr;
                }
            }
            if ( j != BUFFSIZE ) break;
        }
    }
    if ( o_verbose )
        fprintf(stdout," - Done                    \n");
/*    if ( o_verbose ) */
/*        fprintf(stdout,"\r%60s\r"," "); */
    close(fd);
    return OK;
}



int wipe_dir(char *dirname)
{
    struct dirent *d;
    DIR *fd;
    struct stat buff;
    char fn[NAME_MAX+PATH_MAX];

    if ( NULL==dirname ) return ERROR;

    if ( o_interactive )
    {
        fprintf(stderr,"Enter directory %s ",dirname);
        if ( !ask_y_n("(y/[n]/a)?") ) return ERROR;
    }

    /* check directory access */
    if ( !o_wipe_slacks && OK != access(dirname,X_OK|R_OK|W_OK))
    {
        if ( !o_force )
        {
            fprintf(stderr,"Cannot enter to directory %s : %s  \t - Skiping\n",dirname,strerror(errno));
            return ERROR;
        }
        if ( OK != chmod(dirname,S_IRUSR|S_IWUSR|S_IXUSR) )
        {
            fprintf(stderr,"Cannot chmod directory %s: %s  \t - Skiping\n",dirname,strerror(errno));
            return ERROR;
        }
    }

    fd=opendir(dirname);
    if ( NULL == fd )
    {
        fprintf(stderr,"Cannot open directory %s: %s  \t - Skiping\n",dirname,strerror(errno));
        return ERROR;
    }

    if ( o_verbose )
    {
        fprintf (stdout, "Entering directory %s\n",dirname);
    }


    while ( NULL != (d=readdir(fd)) )
    {
        sprintf(fn,"%s/%s",dirname,d->d_name);
        if ( OK != lstat(fn,&buff) )
        {
            fprintf(stderr,"Cannot stat %s: %s  \t - Skiping\n",fn,strerror(errno));
            continue;
        }
        if ( S_ISDIR(buff.st_mode) )    /*directory */
        {
            if ( 0==strcmp(d->d_name,".") ||
                 0==strcmp(d->d_name,"..") )
            {

            }
            else
            {
                wipe_dir(fn);
            }
        }
        else if ( S_ISREG(buff.st_mode) )   /*regular file */
        {
            (void) (wipe_file(fn) || wipe_filename(fn));
        }
        else if ( S_ISLNK(buff.st_mode) )   /*link */
        {
            (void) wipe_filename(fn);
        }
        else if ( S_ISCHR(buff.st_mode) )   /*character device */
        {
            if ( o_wipe_dev ) (void) wipe_file(fn);
            else (void) wipe_filename(fn);
        }
        else if ( S_ISBLK(buff.st_mode) )   /*block device */
        {
            if ( o_wipe_dev ) (void) wipe_file(fn);
            else (void) wipe_filename(fn);
        }
        else if ( S_ISFIFO(buff.st_mode) )  /*named pipe (fifo) */
        {
            (void) wipe_filename(fn);
        }
        else if ( S_ISSOCK(buff.st_mode) )  /*socket */
        {
            (void) wipe_filename(fn);
        }
        else                                /*somthing else .... */
        {
            fprintf(stderr,"Unknown type of file: %s\n",fn);
        }

    }
    closedir(fd);
    if ( o_dont_delete )return OK;

    if ( o_interactive )
    {
        fprintf(stderr,"Remove directory %s ",dirname);
        if ( !ask_y_n("(y/[n]/a)?") ) return ERROR;
    }

    if ( o_verbose )
    {
        fprintf (stdout, "Removing directory %s ",dirname);
    }
    if ( OK != rmdir(dirname) )
    {
        if ( o_verbose )
            fprintf (stderr, "\t - %s\n",strerror(errno));
        else
            fprintf (stderr, "Removing %s - %s\n",dirname,strerror(errno));
        return ERROR;
    }
    if ( o_verbose )
        fprintf (stdout, "\t - Done\n");
    return OK;
}

#if 0
int wipe_free_space(char *path)
{
    size_t free_size=0;
    struct statfs st_buff;
    
    if (OK != statfs(path,&st_buff))
    {
        fprintf(stderr,"statfs(%s,&buff): %s\n",path,strerror(errno));
        return ERROR;
    }

    free_size = st_buff.f_bfree * S_BLKSIZE; //free size in bytes

    return OK;
}
#endif //0

int main(int argc, char* argv[])
{
    struct stat buff;
    int c;
    int i;
    int m_count = 0;

    setvbuf(stderr,(char*)NULL,_IONBF,0);
    setvbuf(stdout,(char*)NULL,_IONBF,0);
    srand(time(NULL));
    while ( (c=getopt(argc,argv,OPTSTRING)) != -1 )
    {
        switch ( (char)c )
        {
            case 'V':
                printf("bcwipe version %s\n",VERSION);
                exit(0);
            case 'v':
                o_verbose=TRUE;
                break;
            case 'r':
                o_recurse=TRUE;
                break;
            case 'f':
                o_force=TRUE;
                break;
            case 'D':
                o_dont_wipe_fn=TRUE;
                break;
            case 'd':
                o_dont_delete=TRUE;
                break;
            case 'b':
                o_wipe_dev=TRUE;
                break;
            case 'i':
                o_interactive=TRUE;
                break;
            case 'I':
                o_interactive=FALSE;
                break;
            case 'm':
                m_count++;
                if (1 == strlen(optarg) && 'd' == optarg[0])
                {
                    o_use_DoD=TRUE;
                    o_pas_num=QPASSES;
                } else
                    if (1 == strlen(optarg) && 'g' == optarg[0])
                {
                    o_use_DoD=FALSE;
                    o_pas_num=DPASSES;
                } else
                {
                    o_pas_num=strtol(optarg, NULL, 0);
                    o_use_DoD=TRUE;
                }
                break;
            case 'u':
                o_user_file=optarg;
                break;
            case 's':
                o_use_rand=TRUE;
                break;
            case 'S':
                o_wipe_slacks=TRUE;
                break;
            case 'p':
                o_use_buff=TRUE;
                break;
            case '?':
            case 'h':
            default:
                fprintf(stdout,usage_string);
                exit(0);
        } /* end of switch */
    } /* end of while */

    argc -= optind;
    argv += optind;

    if ( 0 == argc )
    {
        fprintf(stdout,usage_string);
        exit(0);
    }
    
    if (1 < m_count)
    {
        fprintf(stderr,"Warning: multiple defenition of wipe mode.\n");
        if (o_use_DoD)
        {
            fprintf(stderr,"Using DoD %d passes mode.\n",o_pas_num);
        }else 
            fprintf(stderr,"Using Peter Gutmann mode.\n");

    }
    if ( o_force ) o_interactive=FALSE;
    if (o_wipe_slacks) o_dont_delete = TRUE;

    for ( i = argc;i;i-- )
    {
        if ( OK != lstat(argv[i-1],&buff) )
        {
            fprintf(stderr,"Cannot stat %s: %s  \t - Skiping\n",argv[i-1],strerror(errno));
            continue;
        }
        if ( S_ISDIR(buff.st_mode) )  /*directory */
        {
            if ( o_recurse ) wipe_dir(argv[i-1]);
            else
            {
                fprintf(stderr,"%s is directory  \t - Skiping\n",argv[i-1]);
            }
        }
        else if ( S_ISREG(buff.st_mode) )   /*regular file */
        {
            (void) (wipe_file(argv[i-1]) || wipe_filename(argv[i-1]));
        }
        else if ( S_ISLNK(buff.st_mode) )   /*link */
        {
            (void) wipe_filename(argv[i-1]);
        }
        else if ( S_ISCHR(buff.st_mode) )   /*character device */
        {
            if ( o_wipe_dev ) (void) wipe_file(argv[i-1]);
            else (void) wipe_filename(argv[i-1]);
        }
        else if ( S_ISBLK(buff.st_mode) )   /*block device */
        {
            if ( o_wipe_dev ) (void) wipe_file(argv[i-1]);
            else (void) wipe_filename(argv[i-1]);
        }
        else if ( S_ISFIFO(buff.st_mode) )  /*named pipe (fifo) */
        {
            (void) wipe_filename(argv[i-1]);
        }
        else if ( S_ISSOCK(buff.st_mode) )  /*socket */
        {
            (void) wipe_filename(argv[i-1]);
        }
        else                                /*something else .... */
        {
            fprintf(stderr,"Unknown type of file: %s\n",argv[i-1]);
        }
    }
    if (o_wipe_slacks)
    {
        fprintf(stdout,"Wiped %d Kb of file slacks\n",(unsigned int)(total_size_of_slacks/1024));
    }
    return 0;
}


